from __future__ import annotations
from typing import Any, Union, Optional
from pathlib import Path
import inspect

import ray

from sinagot.item import Item
from sinagot.base import StepBase, SeedBase
from sinagot.logger import get_logger
from sinagot.storage import Storage

logger = get_logger()


class NodeMixin:
    name: str

    def get_storage(self, item: Item) -> Optional[Storage]:
        store_path = getattr(item.workspace, self.name, None)
        if store_path is not None:
            path = item._resolve_path(store_path)
            data_type = item.__annotations__[self.name]
            return Storage(
                path=path,
                data_type=data_type,
                read_kwargs=store_path.read_kwargs,
                write_kwargs=store_path.write_kwargs,
            )
        return None


class Step(StepBase, NodeMixin):
    def __init__(self, func: Any, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.store_path: Union[None, Path] = None

    def __get__(self, item: Item, objtype: Any = None) -> Any:
        return self.get_data(item)

    def get_data(self, item: Item) -> ray.ObjectRef:

        storage = self.get_storage(item)
        if (storage is not None) and storage.exists():
            logger.info("Getting %s data for item %s from storage", self, item.item_id)
            return storage.read()
        else:
            logger.info("Processing %s data for item %s", self, item.item_id)
            func = self.func
            kwargs = {}
            parameters = inspect.signature(func).parameters
            if "item_id" in parameters:
                kwargs["item_id"] = item.item_id

            result = item._run(self, **kwargs)

            if storage is not None:
                logger.info("Saving %s data for item %s", self, item.item_id)
                storage.write(ray.get(result))  # type: ignore

            return ray.get(result)  # type: ignore


def step(func: Any) -> Any:
    def step(*args: Any, **kwargs: Any) -> Step:
        return Step(func, *args, **kwargs)

    func.step = step
    return func


class Seed(SeedBase, NodeMixin):
    def __get__(self, item: Item, objtype: Any = None) -> Any:
        return self.get_data(item)

    def get_data(self, item: Item) -> ray.ObjectRef:
        storage = self.get_storage(item)
        if storage is not None:
            logger.info("Getting %s data for item %s", self, item.item_id)
            return storage.read()


def seed() -> Any:
    return Seed()



import pkg_resources

from .item import Item
from .workspace import Workspace
from .node import step, seed
from .storage import LocalStorage

__all__ = ["local_file", "step", "Workspace", "Item", "seed", "LocalStorage"]

__version__ = pkg_resources.get_distribution("sinagot").version



from __future__ import annotations
from typing import Union, Any, Dict, Type, Optional
from pathlib import Path
import os

import pandas as pd

from sinagot.base import AttrBase
from sinagot.logger import get_logger

logger = get_logger()


class LocalStorage(AttrBase):
    def __init__(
        self,
        path: Union[str, Path],
        read_kwargs: Optional[Dict[str, Any]] = None,
        write_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.path = path
        self.read_kwargs = read_kwargs or {}
        self.write_kwargs = write_kwargs or {}

    def __fspath__(self) -> str:
        return os.fspath(self.path)


class Storage:
    def __init__(
        self,
        path: Path,
        data_type: Any,
        read_kwargs: Dict[str, Any],
        write_kwargs: Dict[str, Any],
    ):
        self.path = path
        self.data_type = data_type
        self.format = self.path.suffix
        self.read_kwargs = read_kwargs
        self.write_kwargs = write_kwargs

    def _get_handler(self) -> TypeHandler:
        try:
            return type_handlers[self.data_type](
                read_kwargs=self.read_kwargs, write_kwargs=self.write_kwargs
            )
        except KeyError:
            raise NotImplementedError(f"{self.data_type.__name__} not implemented")

    def exists(self) -> bool:
        return self.path.exists()

    def read(
        self,
    ) -> Any:
        return self._get_handler().read(self.path)

    def write(self, data: Any) -> None:
        self._get_handler().write(self.path, data)


type_handlers: Dict[Any, Type[TypeHandler]] = {}


class TypeHandler:
    def __init_subclass__(cls, data_type: Any) -> None:
        type_handlers[data_type] = cls

    def __init__(
        self,
        read_kwargs: Dict[str, Any],
        write_kwargs: Dict[str, Any],
    ):
        self.read_kwargs = read_kwargs
        self.write_kwargs = write_kwargs

    def read(self, path: Path) -> Any:
        """read"""

    def write(self, path: Path, data: Any) -> None:
        """write"""


class StringHandler(TypeHandler, data_type=str):
    def read(self, path: Path) -> Any:
        return path.read_text(**self.read_kwargs)

    def write(self, path: Path, data: Any) -> None:
        path.write_text(data, **self.write_kwargs)


class IntHandler(TypeHandler, data_type=int):
    def read(self, path: Path) -> Any:
        return int(path.read_text(**self.read_kwargs))

    def write(self, path: Path, data: Any) -> None:
        path.write_text(str(data), **self.write_kwargs)


class DataFrameHandler(TypeHandler, data_type=pd.DataFrame):
    def read(self, path: Path) -> Any:
        return pd.read_csv(path, **self.read_kwargs)

    def write(self, path: Path, data: Any) -> None:
        data.to_csv(path, **self.write_kwargs)



from __future__ import annotations
from pathlib import Path
from glob import glob
import re
from typing import (
    Iterator,
    Generator,
    Union,
    Type,
    Generic,
    TypeVar,
    Mapping,
)

from sinagot.base import ItemBase
from sinagot.path_template import PathTemplate
from sinagot.logger import get_logger
from sinagot.config import get_settings
from sinagot.ray import init_ray

logger = get_logger()

IT = TypeVar("IT", bound=ItemBase)


class Workspace(Generic[IT], Mapping[str, IT]):

    ITEM_RE = ".*"
    Item: Type[IT]

    def __init__(self, *root_path_segments: Union[str, Path]):
        self.root_path = Path(*root_path_segments)
        self.settings = get_settings()
        init_ray()
        if type(self) != Workspace:
            logger.info("%s initialized", str(self))
            logger.debug("%s config :  %s", str(self), self.settings)

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self.root_path}')"

    @property
    def item_class(self) -> Type[IT]:
        return self.__orig_bases__[0]().__orig_class__.__args__[0]  # type: ignore

    def __getitem__(self, item_id: str) -> IT:
        return self.item_class(self, item_id)  # type: ignore

    def __iter__(self) -> Iterator[str]:
        yield from self._iter_item_ids()

    def _iter_item_ids(self) -> Generator[str, None, None]:
        item_ids = set()
        data = [
            self._resolve_path(getattr(self, name)) for name in self.item_class._seeds
        ]
        glob_patterns = [item.format(item_id="*") for item in data]
        re_patterns = [str(item.format(item_id=f"({self.ITEM_RE})")) for item in data]
        for glob_pattern, re_pattern in zip(glob_patterns, re_patterns):
            for file_str in glob(str(glob_pattern)):
                item_match = re.match(re_pattern, file_str)
                if item_match:
                    item_id = item_match.group(1)
                    if item_id not in item_ids:
                        item_ids.add(item_id)
                        yield item_id

    def __len__(self) -> int:
        return len(list(self._iter_item_ids()))

    def _resolve_path(self, path: Path, **kwargs: str) -> PathTemplate:
        return PathTemplate(self.root_path, path).format(**kwargs)



from typing import Type

import networkx as nx

from sinagot.base import ItemBase


class Workflow:
    graph: nx.Graph

    def __init__(self, item: Type[ItemBase]):
        self.item = item
        self._set_graph()

    def _set_graph(self) -> None:
        item = self.item
        graph = nx.DiGraph()
        graph.add_nodes_from(item._seeds)
        graph.add_nodes_from(item._steps)
        for name, step in item._steps.items():
            graph.add_edges_from(
                [
                    (source.name, name, {"label": step.func.__name__})
                    for source in step.args
                ]
            )
            graph.add_edges_from(
                [
                    (source.name, name, {"label": step.func.__name__})
                    for source in step.kwargs.values()
                ]
            )
        self.graph = graph

    def draw(self) -> None:
        graph = self.graph

        pos = nx.nx_pydot.pydot_layout(graph, prog="dot")
        edge_labels = {
            (source, dest): data["label"] for source, dest, data in graph.edges.data()
        }

        nx.draw_networkx(
            graph,
            pos=pos,
            arrows=True,
            node_color="#e1a798ff",
            font_color="#542418",
            font_weight="bold",
            node_shape="8",
            edge_color="#1a2127",
        )
        nx.draw_networkx_edge_labels(
            graph, pos=pos, edge_labels=edge_labels, rotate=False
        )
